from typing import NamedTuple
from datetime import datetime
from .constants import *

"""bot-dev-constructed objects"""


class Reply:
    """Helper class to be constructed by the bot developer.
        Avoids quirky dict syntax because ``reply.key``
        is so much more succint than ``dict["key"]``.
    """

    def __init__(self, text: str, visibility=None, spoiler_text=None):
        if not text:
            raise ValueError("Argument `text` cannot be empty")
        self.text = str(text)

        if not visibility in VISIBILITY_LIST + [None]:
            raise ValueError(
                "Argument `visibility` must be one of "
                "'public', 'unlisted', 'private', or 'direct'"
            )
        self.visibility = visibility

        self.spoiler_text = spoiler_text


# return Boost or return Favourite in callbacks
Boost = object()
Favourite = object()

"""mastobot objects provided for bot dev"""

Mention = NamedTuple(
    "Account", [("id", int), ("username", str), ("acct", str), ("url", str)]
)

Account = NamedTuple(
    "Account",
    [
        ("id", int),
        ("username", str),
        ("acct", str),
        ("display_name", str),
        ("url", str),
        ("avatar", str),
        ("avatar_static", str),
        ("bot", bool),
    ],
)

Status = NamedTuple(
    "Status",
    [
        ("id", int),
        ("url", str),
        ("account", Account),
        ("in_reply_to_id", int),
        ("content", str),
        ("created_at", datetime),
        ("sensitive", bool),
        ("spoiler_text", str),
        ("visibility", str),
        ("mentions", list),
    ],
)

Notification = NamedTuple(
    "Notification",
    [
        ("id", int),
        ("type", str),
        ("created_at", datetime),
        ("account", Account),
        ("status", Status),
    ],
)


def gen_status(obj: dict) -> Status:
    """Dict to NamedTuple utility.

    :param obj: (dict) mastodon.py status dict.
    """
    return Status(
        id=obj["id"],
        url=obj["url"],
        account=Account(
            id=obj["account"]["id"],
            username=obj["account"]["username"],
            acct=obj["account"]["acct"],
            display_name=obj["account"]["display_name"],
            url=obj["account"]["url"],
            avatar=obj["account"]["avatar"],
            avatar_static=obj["account"]["avatar_static"],
            bot=obj["account"]["bot"],
        ),
        in_reply_to_id=obj["in_reply_to_id"],
        content=obj["content"],
        created_at=obj["created_at"],
        sensitive=obj[SENSITIVE],
        spoiler_text=obj[SPOILER_TEXT],
        visibility=obj[VISIBILITY],
        mentions=[
            Mention(id=m["id"], username=m["username"], acct=m["acct"], url=m["url"])
            for m in obj["mentions"]
        ],
    )


def gen_account(obj: dict) -> Account:
    """Dict to NamedTuple utility.

    :param obj: (dict) mastodon.py user dict.
    """

    return Account(
        id=obj["id"],
        username=obj["username"],
        acct=obj["acct"],
        display_name=obj["display_name"],
        url=obj["url"],
        avatar=obj["avatar"],
        avatar_static=obj["avatar_static"],
        bot=obj["bot"],
    )



# List common constants used by Mastobot and Mastodon.py here
# Comes handy if you have autocomplete

# Mastobot section
# StreamListener.on_xxx()
UPDATE = "update"
NOTIFICATION = "notification"

# Trigger(validation="xxx")
EQUALS = "equals"
CONTAINS = "contains"
REGEX = "regex"
EVALUATE = "evaluate"
VALIDATION_LIST = [EQUALS, CONTAINS, REGEX, EVALUATE]

# Mastodon.py section
# toot = {"xxx": "..."}
STATUS = "status"
SENSITIVE = "sensitive"
VISIBILITY = "visibility"
SPOILER_TEXT = "spoiler_text"

# toot = {"visibility": "xxx"}
PUBLIC = "public"
UNLISTED = "unlisted"
PRIVATE = "private"
DIRECT = "direct"
VISIBILITY_LIST = [PUBLIC, UNLISTED, PRIVATE, DIRECT]

# notification= {"type": "xxx"}
MENTION = "mention"
FAVOURITE = "favourite"
REBLOG = "reblog"
FOLLOW = "follow"
UPDATE = "update"
EVENT_LIST = [MENTION, FAVOURITE, REBLOG, FOLLOW, UPDATE]




from setuptools import setup, find_packages

setup(
    name="mastobot",
    version="0.2.0",
    description="Create cheap Mastodon bots in a Flask-like syntax",
    long_description=open("./README.md").read(),
    long_description_content_type="text/markdown",
    keywords="mastodon bot",
    license="0BSD",
    url="https://git.sr.ht/~fkfd/mastobot",
    author="Frederick Yin",
    author_email="fkfd@macaw.me",
    classifiers=["Development Status :: 3 - Alpha", "License :: Public Domain",],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["Mastodon.py", "websockets", "lxml"],
)



import websockets
import asyncio
import urllib
import json
from .constants import *


class WebsocketListener:
    def __init__(self, instance: dict, access_token: str, stream="user"):
        """Initiate a WebSocket Listener

        :param instance: (dict) Mastodon.py dict of the instance
        :param access_token: (str)
        :param stream: (str)
        """

        self._instance = instance["uri"]
        try:
            self._ws_endpoint = urllib.parse.urljoin(
                instance["urls"]["streaming_api"],
                f"/api/v1/streaming?access_token={access_token}&stream={stream}",
            )
        except KeyError:
            raise Exception(
                f"The instance {instance['uri']} did not provide a WebSocket URL"
            )

    def on_update(self, status):
        """Stub method to be overwritten by bot dev"""
        pass

    def on_notification(self, notification):
        """Stub method to be overwritten by bot dev"""
        pass

    async def _stream(self):
        async with websockets.connect(self._ws_endpoint) as socket:
            while True:
                received = await socket.recv()
                # received:
                # {
                #    "event": "update|notification|...",
                #    "payload": "{\"account\":{...}}"
                # }
                # payload is a string of a JSON object, not one itself
                try:
                    response = json.loads(received)
                    event = response["event"]
                    payload = json.loads(response["payload"])
                    if event == UPDATE:
                        self.on_update(payload)
                    elif event == NOTIFICATION:
                        self.on_notification(payload)
                except json.decoder.JSONDecodeError:
                    # occasionally the server would send an empty event
                    pass

    def start_stream(self):
        asyncio.get_event_loop().run_until_complete(self._stream())

